with Interfaces.C.Strings; use Interfaces.C, Interfaces.C.Strings;
with System;               use System;

package body Erl_Interface.Erl_Connect is

   type C_Emessage is record
      Typ     : int;
      Msg     : Eterm_Ptr;
      To      : Eterm_Ptr;
      From    : Eterm_Ptr;
      To_Name : char_array (1 .. 128);   --  Bound to be checked ???
   end record;
   pragma Convention (C, C_Emessage);

   procedure Copy (Source : access C_Emessage; Destination : access Emessage);
   --  Copy a C message to an Ada message and free the associated memory

   ----------
   -- Copy --
   ----------

   procedure Copy (Source : access C_Emessage; Destination : access Emessage)
   is
      procedure free (Add : in Address);
      pragma Import (C, free, "free");
   begin
      Destination.Kind    := Message_Kind'Val (Source.Typ);
      Destination.Msg     := Source.Msg;
      Destination.To      := Source.To;
      Destination.From    := Source.From;
      Destination.To_Name := new String'(To_Ada (Source.To_Name));
      free (Source.all'Address);
   end Copy;

   -----------------
   -- Erl_Connect --
   -----------------

   function Erl_Connect (Node : String) return File_Descriptor is
      function erl_connect (Node : Address) return int;
      pragma Import (C, erl_connect, "erl_connect");
      FD : constant int := erl_connect (To_C (Node)'Address);
   begin
      if FD < 0 then
         raise Erl_Error;
      else
         return File_Descriptor (FD);
      end if;
   end Erl_Connect;

   ----------------------
   -- Erl_Connect_Init --
   ----------------------

   procedure Erl_Connect_Init
     (Number   : in Integer;
      Cookie   : in String;
      Creation : in Integer)
   is
      function erl_connect_init
        (Number   : int;
         Cookie   : Address;
         Creation : short)
        return int;
      pragma Import (C, erl_connect_init, "erl_connect_init");
   begin
      if erl_connect_init
        (int (Number), To_C (Cookie)'Address, short (Creation)) = 0
      then
         raise Erl_Error;
      end if;
   end Erl_Connect_Init;

   -----------------------
   -- Erl_Connect_Xinit --
   -----------------------

   procedure Erl_Connect_Xinit
     (Host     : in String;
      Alive    : in String;
      Node     : in String;
      Addr     : in In_Addr;
      Cookie   : in String;
      Creation : in Integer)
   is
      function erl_connect_xinit
        (Host     : Address;
         Alive    : Address;
         Node     : Address;
         Addr     : In_Addr;
         Cookie   : Address;
         Creation : short)
        return int;
      pragma Import (C, erl_connect_xinit, "erl_connect_xinit");
      C_Host   : char_array := To_C (Host);
      C_Alive  : char_array := To_C (Alive);
      C_Node   : char_array := To_C (Node);
      C_Cookie : char_array := To_C (Cookie);
   begin
      if erl_connect_xinit (C_Host'Address, C_Alive'Address, C_Node'Address,
                            Addr, C_Cookie'Address, short (Creation)) = 0 then
         raise Erl_Error;
      end if;
   end Erl_Connect_Xinit;

   --------------------------
   -- Erl_Close_Connection --
   --------------------------

   procedure Erl_Close_Connection (FD : in out File_Descriptor) is
      function erl_close_connection (FD : int) return int;
      pragma Import (C, erl_close_connection, "erl_close_connection");
   begin
      if erl_close_connection (int (FD)) /= 0 then
         raise Erl_Error;
      end if;
      FD := -1;
   end Erl_Close_Connection;

   -----------------
   -- Erl_Publish --
   -----------------

   function Erl_Publish (Port : Unsigned_16) return File_Descriptor is
      function erl_publish (Port : short) return int;
      pragma Import (C, erl_publish, "erl_publish");
      FD : constant int := erl_publish (short (Port));
   begin
      if FD = -1 then
         raise Erl_Error;
      else
         return File_Descriptor (FD);
      end if;
   end Erl_Publish;

   -----------------
   -- Erl_Receive --
   -----------------

   function Erl_Receive
     (FD     : File_Descriptor;
      Buffer : access Buffer_Type)
     return Natural
   is
      function erl_receive (FD : int; Buffer : Address; Size : int) return int;
      pragma Import (C, erl_receive, "erl_receive");
      Read : constant int := erl_receive (int (FD),
                                          Buffer.all'Address,
                                          Buffer'Length);
   begin
      if Read < 0 then
         raise Erl_Error;
      else
         return Natural (Read);
      end if;
   end Erl_Receive;

   ---------------------
   -- Erl_Receive_Msg --
   ---------------------

   function Erl_Receive_Msg
     (FD      : File_Descriptor;
      Buffer  : access Buffer_Type;
      Message : access Emessage)
     return Integer
   is
      function erl_receive_msg
        (FD : int; Buffer : Address; Size : int; Message : access C_Emessage)
        return int;
      pragma Import (C, erl_receive_msg, "erl_receive_msg");
      C_Msg : aliased C_Emessage;
      Code  : constant int := erl_receive_msg (int (FD),
                                               Buffer.all'Address,
                                               Buffer'Length,
                                               C_Msg'Access);
   begin
      if Code < 0 then
         raise Erl_Error;
      else
         Copy (C_Msg'Access, Message);
         return Integer (Code);
      end if;
   end Erl_Receive_Msg;

   ------------------
   -- Erl_Reg_Send --
   ------------------

   procedure Erl_Reg_Send
     (FD  : in File_Descriptor;
      To  : in String;
      Msg : access Eterm)
   is
      function erl_reg_send (FD : int; To : Address; Msg : access Eterm)
        return int;
      pragma Import (C, erl_reg_send, "erl_reg_send");
   begin
      if erl_reg_send (int (FD), To_C (To)'Address, Msg) = 0 then
         raise Erl_Error;
      end if;
   end Erl_Reg_Send;

   -------------
   -- Erl_RPC --
   -------------

   function Erl_RPC
     (FD     : File_Descriptor;
      Module : String;
      Func   : String;
      Args   : access Eterm)
     return Eterm_Ptr
   is
      function erl_rpc
        (FD     : int;
         Module : Address;
         Func   : Address;
         Args   : access Eterm)
        return Eterm_Ptr;
      pragma Import (C, erl_rpc, "erl_rpc");
      Result : constant Eterm_Ptr :=
        erl_rpc (int (FD), To_C (Module)'Address,
                 To_C (Func)'Address, Args);
   begin
      if Result = null then
         raise Erl_Error;
      else
         return Result;
      end if;
   end Erl_RPC;

   ------------------
   -- Erl_RPC_From --
   ------------------

   procedure Erl_RPC_From
     (FD      : in File_Descriptor;
      Timeout : in Duration;
      Emsg    : access Emessage)
   is
      function erl_rpc_from
        (FD : int; Timeout : int; Emsg : access C_Emessage)
        return int;
      pragma Import (C, erl_rpc_from, "erl_rpc_from");
      C_Emsg : aliased C_Emessage;
      Code   : constant int :=
        erl_rpc_from (int (FD), int (Timeout * 1000), C_Emsg'Access);
   begin
      if Code = int (Erl_Msg) then
         Copy (C_Emsg'Access, Emsg);
      elsif Code = -1 then  --  Code to check
         raise Erl_Timeout;
      else
         raise Erl_Error;
      end if;
   end Erl_RPC_From;

   ----------------
   -- Erl_RPC_To --
   ----------------

   procedure Erl_RPC_To
     (FD     : File_Descriptor;
      Module : String;
      Func   : String;
      Args   : access Eterm)
   is
      procedure erl_rpc_to
        (FD     : in int;
         Module : in Address;
         Func   : in Address;
         Args : access Eterm);
      pragma Import (C, erl_rpc_to, "erl_rpc_to");
   begin
      erl_rpc_to (int (FD), To_C (Module)'Address, To_C (Func)'Address, Args);
   end Erl_RPC_To;

   --------------
   -- Erl_Send --
   --------------

   procedure Erl_Send
     (FD  : in File_Descriptor;
      To  : access Eterm;
      Msg : access Eterm)
   is
      function erl_send
        (FD : int; To : Address; Msg : access Eterm)
        return int;
      pragma Import (C, erl_send, "erl_send");
   begin
      if erl_send (int (FD), To.all'Address, Msg) = 0 then
         raise Erl_Error;
      end if;
   end Erl_Send;

   ------------------
   -- Erl_Xconnect --
   ------------------

   function Erl_Xconnect
     (Addr : In_Addr; Alive : String)
     return File_Descriptor
   is
      function erl_xconnect (Addr : In_Addr; Alive : Address) return int;
      pragma Import (C, erl_xconnect, "erl_xconnect");
      FD : constant int := erl_xconnect (Addr, To_C (Alive)'Address);
   begin
      if FD < 0 then
         raise Erl_Error;
      else
         return File_Descriptor (FD);
      end if;
   end Erl_Xconnect;

end Erl_Interface.Erl_Connect;
