with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;
with Interfaces.C.Strings;       use Interfaces.C, Interfaces.C.Strings;
with System;                     use System;

package body Erl_Interface.Erl_Format is

   procedure Free is new Ada.Unchecked_Deallocation
     (Stream_Element_Array,
      Stream_Element_Access);

   function To_Int is new Ada.Unchecked_Conversion (chars_ptr, int);
   function To_Int is new Ada.Unchecked_Conversion (double, int);
   --  ??? The last one is clearly wrong

   type Argument_Array is array (1 .. 10) of int;

   function erl_format
     (Format_String : Address;
      Arg1          : int;
      Arg2          : int;
      Arg3          : int;
      Arg4          : int;
      Arg5          : int;
      Arg6          : int;
      Arg7          : int;
      Arg8          : int;
      Arg9          : int;
      Arg10         : int)
     return Eterm_Ptr;
   pragma Import (C, erl_format, "erl_format");

   function Format
     (Stream          : access Format_Stream;
      Arguments       : access Argument_Array;
      Format_Index    : Positive;
      Arguments_Index : Positive)
     return Eterm_Ptr;
   --  Apply the format string. Note that, except when resources must be
   --  released after the call (in case of C strings notably), this function
   --  is tail recursive.

   -------------
   -- Compute --
   -------------

   function Compute (Stream : access Format_Stream) return Eterm_Ptr is
      Arguments : aliased Argument_Array;
   begin
      return Format (Stream, Arguments'Access,
                     Stream.Format_String'First, Arguments'First);
   end Compute;

   ------------
   -- Format --
   ------------

   function Format
     (Stream          : access Format_Stream;
      Arguments       : access Argument_Array;
      Format_Index    : Positive;
      Arguments_Index : Positive)
     return Eterm_Ptr
   is
      I : Natural := Format_Index;
   begin
      while I <= Stream.Format_String'Last - 1 loop
         if Stream.Format_String (I) = '~' then
            case Stream.Format_String (I + 1) is

               when 'i' =>
                  Arguments (Arguments_Index) := int (Integer'Input (Stream));
                  return
                    Format (Stream, Arguments, I + 2, Arguments_Index + 1);

               when 'f' =>
                  Arguments (Arguments_Index) :=
                    To_Int (double (Float'Input (Stream)));
                  return
                    Format (Stream, Arguments, I + 2, Arguments_Index + 1);

               when 'a' | 's' =>
                  declare
                     C_Str  : chars_ptr := New_String (String'Input (Stream));
                     Result : Eterm_Ptr;
                  begin
                     Arguments (Arguments_Index) := To_Int (C_Str);
                     Result :=
                        Format (Stream, Arguments, I + 2, Arguments_Index + 1);
                     Free (C_Str);
                     return Result;
                  end;

               when 'w' =>
                  Arguments (Arguments_Index) := int'Input (Stream);
                  return
                    Format (Stream, Arguments, I + 2, Arguments_Index + 1);

               when '~' =>
                  I := I + 1;

               when others =>
                  raise Format_Error;

            end case;
         end if;

         I := I + 1;
      end loop;

      return erl_format (To_C (Stream.Format_String.all)'Address,
                         Arguments (1),
                         Arguments (2),
                         Arguments (3),
                         Arguments (4),
                         Arguments (5),
                         Arguments (6),
                         Arguments (7),
                         Arguments (8),
                         Arguments (9),
                         Arguments (10));
   end Format;

   ----------------
   -- Initialize --
   ----------------

   procedure Initialize
     (Stream        : access Format_Stream;
      Format_String : in String)
   is
   begin
      Stream.Format_String := new String'(Format_String);
   end Initialize;

   -----------
   -- Match --
   -----------

   function Match
     (Pattern : access Eterm; Term : access Eterm)
     return Boolean
   is
      function erl_match (Pattern : access Eterm; Term : access Eterm)
        return int;
      pragma Import (C, erl_match, "erl_match");
   begin
      return erl_match (Pattern, Term) /= 0;
   end Match;

   ----------
   -- Read --
   ----------

   procedure Read
     (Stream : in out Format_Stream;
      Item   : out Stream_Element_Array;
      Last   : out Stream_Element_Offset)
   is
      Rest : Stream_Element_Count;
   begin
      if Stream.Bytes = null then
         Last := Item'First - 1;
         return;
      end if;

      Rest := Stream.Bytes'Last - Stream.Index + 1;

      if Rest <= Item'Length then
         Last := Item'First + Rest - 1;
         Item (Item'First .. Last) :=
           Stream.Bytes (Stream.Index .. Stream.Bytes'Last);
         Free (Stream.Bytes);
      else
         Item := Stream.Bytes (Stream.Index .. Stream.Index + Item'Length - 1);
         Stream.Index := Stream.Index + Item'Length;
         Last := Item'Last;
      end if;
   end Read;

   -----------
   -- Write --
   -----------

   procedure Write
     (Stream : in out Format_Stream;
      Item   : in Stream_Element_Array)
   is
      Old : Stream_Element_Access := Stream.Bytes;
   begin
      if Old = null then
         Stream.Bytes := new Stream_Element_Array'(Item);
         Stream.Index := Stream.Bytes'First;
      else
         Stream.Bytes := new Stream_Element_Array'(Stream.Bytes.all & Item);
         Free (Old);
      end if;
   end Write;

end Erl_Interface.Erl_Format;
