with Ada.Streams;             use Ada.Streams;
with Erl_Interface.Erl_Eterm; use Erl_Interface.Erl_Eterm;

package Erl_Interface.Erl_Format is

   Format_Error : exception;

   type Format_Stream is new Root_Stream_Type with private;

   procedure Read
     (Stream : in out Format_Stream;
      Item   : out Stream_Element_Array;
      Last   : out Stream_Element_Offset);

   procedure Write
     (Stream : in out Format_Stream;
      Item   : in Stream_Element_Array);

   procedure Initialize
     (Stream        : access Format_Stream;
      Format_String : in String);
   --  Initialize the format stream

   function Compute (Stream : access Format_Stream) return Eterm_Ptr;
   --  Return an Eterm corresponding to the format string and the data
   --  entered into the stream. A new initialize is required to pursue
   --  using the stream. Can raise Format_Error.

   function Match
     (Pattern : access Eterm; Term : access Eterm)
     return Boolean;
   --  Return True if Pattern matches Term. If free variables were present
   --  in Pattern, their value may be retrieved using Var_Content.

private

   type String_Access is access String;

   type Stream_Element_Access is access Stream_Element_Array;

   type Format_Stream is new Root_Stream_Type with record
      Format_String : String_Access;
      Bytes         : Stream_Element_Access;
      Index         : Stream_Element_Offset;
   end record;

end Erl_Interface.Erl_Format;
